from typing import Optional, Tuple, Type

from .capsule import Capsule
from .curve_point import CurvePoint
from .curve_scalar import CurveScalar
from .errors import VerificationError
from .hashing import hash_to_cfrag_verification, kfrag_signature_message
from .keys import PublicKey
from .key_frag import KeyFrag, KeyFragID
from .params import PARAMETERS
from .serializable import Serializable, Deserializable, HasSerializedSize
from .signing import Signature


class CapsuleFragProof(Serializable, Deserializable):

    def __init__(self,
                 point_e2: CurvePoint,
                 point_v2: CurvePoint,
                 kfrag_commitment: CurvePoint,
                 kfrag_pok: CurvePoint,
                 signature: CurveScalar,
                 kfrag_signature: Signature,
                 ):

        self.point_e2 = point_e2
        self.point_v2 = point_v2
        self.kfrag_commitment = kfrag_commitment
        self.kfrag_pok = kfrag_pok
        self.signature = signature
        self.kfrag_signature = kfrag_signature

    def _components(self):
        return (self.point_e2, self.point_v2, self.kfrag_commitment,
                self.kfrag_pok, self.signature, self.kfrag_signature)

    _COMPONENT_TYPES: Tuple[Type[HasSerializedSize], ...] = (
        CurvePoint, CurvePoint, CurvePoint, CurvePoint, CurveScalar, Signature)
    _SERIALIZED_SIZE = sum(tp.serialized_size() for tp in _COMPONENT_TYPES)

    def __eq__(self, other):
        return self._components() == other._components()

    @classmethod
    def serialized_size(cls):
        return cls._SERIALIZED_SIZE

    @classmethod
    def _from_exact_bytes(cls, data):
        return cls(*cls._split(data, *cls._COMPONENT_TYPES))

    def __bytes__(self):
        return b''.join(bytes(comp) for comp in self._components())

    @classmethod
    def from_kfrag_and_cfrag(cls,
                             capsule: Capsule,
                             kfrag: KeyFrag,
                             cfrag_e1: CurvePoint,
                             cfrag_v1: CurvePoint,
                             ) -> 'CapsuleFragProof':

        params = PARAMETERS

        rk = kfrag.key
        t = CurveScalar.random_nonzero()

        # Here are the formulaic constituents shared with `CapsuleFrag.verify()`.

        e = capsule.point_e
        v = capsule.point_v

        e1 = cfrag_e1
        v1 = cfrag_v1

        u = params.u
        u1 = kfrag.proof.commitment

        e2 = e * t
        v2 = v * t
        u2 = u * t

        h = hash_to_cfrag_verification([e, e1, e2, v, v1, v2, u, u1, u2])

        ###

        z3 = t + rk * h

        return cls(point_e2=e2,
                   point_v2=v2,
                   kfrag_commitment=u1,
                   kfrag_pok=u2,
                   signature=z3,
                   kfrag_signature=kfrag.proof.signature_for_receiver,
                   )


class CapsuleFrag(Serializable, Deserializable):
    """
    Re-encrypted fragment of :py:class:`Capsule`.
    """

    def __init__(self,
                 point_e1: CurvePoint,
                 point_v1: CurvePoint,
                 kfrag_id: KeyFragID,
                 precursor: CurvePoint,
                 proof: CapsuleFragProof,
                 ):

        self.point_e1 = point_e1
        self.point_v1 = point_v1
        self.kfrag_id = kfrag_id
        self.precursor = precursor
        self.proof = proof

    def _components(self):
        return (self.point_e1, self.point_v1, self.kfrag_id, self.precursor, self.proof)

    _COMPONENT_TYPES: Tuple[Type[HasSerializedSize], ...] = (
        CurvePoint, CurvePoint, KeyFragID, CurvePoint, CapsuleFragProof)
    _SERIALIZED_SIZE = sum(tp.serialized_size() for tp in _COMPONENT_TYPES)

    def __eq__(self, other):
        return self._components() == other._components()

    def __hash__(self):
        return hash((self.__class__, bytes(self)))

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"

    @classmethod
    def serialized_size(cls):
        return cls._SERIALIZED_SIZE

    @classmethod
    def _from_exact_bytes(cls, data):
        return cls(*cls._split(data, *cls._COMPONENT_TYPES))

    def __bytes__(self):
        return b''.join(bytes(comp) for comp in self._components())

    @classmethod
    def reencrypted(cls, capsule: Capsule, kfrag: KeyFrag) -> 'CapsuleFrag':
        rk = kfrag.key
        e1 = capsule.point_e * rk
        v1 = capsule.point_v * rk
        proof = CapsuleFragProof.from_kfrag_and_cfrag(capsule, kfrag, e1, v1)

        return cls(point_e1=e1,
                   point_v1=v1,
                   kfrag_id=kfrag.id,
                   precursor=kfrag.precursor,
                   proof=proof,
                   )

    def verify(self,
               capsule: Capsule,
               verifying_pk: PublicKey,
               delegating_pk: PublicKey,
               receiving_pk: PublicKey,
               ) -> 'VerifiedCapsuleFrag':
        """
        Verifies the validity of this fragment.
        """

        params = PARAMETERS

        # Here are the formulaic constituents shared with
        # `CapsuleFragProof.from_kfrag_and_cfrag`.

        e = capsule.point_e
        v = capsule.point_v

        e1 = self.point_e1
        v1 = self.point_v1

        u = params.u
        u1 = self.proof.kfrag_commitment

        e2 = self.proof.point_e2
        v2 = self.proof.point_v2
        u2 = self.proof.kfrag_pok

        h = hash_to_cfrag_verification([e, e1, e2, v, v1, v2, u, u1, u2])

        ###

        precursor = self.precursor
        kfrag_id = self.kfrag_id

        kfrag_message = kfrag_signature_message(kfrag_id=self.kfrag_id,
                                                commitment=self.proof.kfrag_commitment,
                                                precursor=self.precursor,
                                                maybe_delegating_pk=delegating_pk,
                                                maybe_receiving_pk=receiving_pk)

        if not self.proof.kfrag_signature.verify(verifying_pk, kfrag_message):
            raise VerificationError("Invalid KeyFrag signature")

        z = self.proof.signature

        # TODO: if one or more of the values here are incorrect,
        # we'll get the wrong `h` (since they're all hashed into it),
        # so perhaps it's enough to check only one of these equations.
        # See https://github.com/nucypher/rust-umbral/issues/46 for details.
        correct_reencryption_of_e = e * z == e2 + e1 * h
        correct_reencryption_of_v = v * z == v2 + v1 * h
        correct_rk_commitment = u * z == u2 + u1 * h

        if not (correct_reencryption_of_e and correct_reencryption_of_v and correct_rk_commitment):
            raise VerificationError("Failed to verify reencryption proof")

        return VerifiedCapsuleFrag(self)


class VerifiedCapsuleFrag(Serializable):
    """
    Verified capsule frag, good for decryption.
    Can be cast to ``bytes``, but cannot be deserialized from bytes directly.
    It can only be obtained from :py:meth:`CapsuleFrag.verify`.
    """

    def __init__(self, cfrag: CapsuleFrag):
        self.cfrag = cfrag

    def __bytes__(self):
        return bytes(self.cfrag)

    @classmethod
    def serialized_size(cls):
        return CapsuleFrag.serialized_size()

    def __eq__(self, other):
        return self.cfrag == other.cfrag

    def __hash__(self):
        return hash((self.__class__, bytes(self)))

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"



from abc import abstractmethod, ABC
from typing import Tuple, Type, List, Any, TypeVar


class HasSerializedSize(ABC):
    """
    A base serialization mixin, denoting a type with a constant-size serialized representation.
    """

    @classmethod
    @abstractmethod
    def serialized_size(cls) -> int:
        """
        Returns the size in bytes of the serialized representation of this object
        (obtained with ``bytes()``).
        """
        raise NotImplementedError


class Deserializable(HasSerializedSize):
    """
    A mixin for composable deserialization.
    """

    Self = TypeVar('Self', bound='Deserializable')

    @classmethod
    def from_bytes(cls: Type[Self], data: bytes) -> Self:
        """
        Restores the object from serialized bytes.
        """
        expected_size = cls.serialized_size()
        if len(data) != expected_size:
            raise ValueError(f"Expected {expected_size} bytes, got {len(data)}")
        return cls._from_exact_bytes(data)

    @staticmethod
    def _split(data: bytes, *types: Type) -> List[Any]:
        """
        Given a list of ``Deserializable`` types, attempts to deserialize them from the bytestring
        one by one and returns the list of the resulting objects and the remaining bytestring.
        """
        objs = []
        pos = 0

        for tp in types:

            if issubclass(tp, bool):
                size = bool_serialized_size()
            else:
                size = tp.serialized_size()

            chunk = data[pos:pos+size]

            if issubclass(tp, bool):
                obj = bool_from_exact_bytes(chunk)
            else:
                obj = tp._from_exact_bytes(chunk)

            objs.append(obj)
            pos += size

        return objs

    @classmethod
    @abstractmethod
    def _from_exact_bytes(cls: Type[Self], data: bytes) -> Self:
        """
        Deserializes the object from a bytestring of exactly the expected length
        (defined by ``serialized_size()``).
        """
        raise NotImplementedError


class Serializable(HasSerializedSize):
    """
    A mixin for composable serialization.
    """

    @abstractmethod
    def __bytes__(self):
        """
        Serializes the object into bytes.
        """
        raise NotImplementedError


def bool_serialized_size() -> int:
    return 1


def bool_bytes(b: bool) -> bytes:
    return b'\x01' if b else b'\x00'


def bool_from_exact_bytes(data: bytes) -> bool:
    if data == b'\x01':
        b = True
    elif data == b'\x00':
        b = False
    else:
        raise ValueError("Incorrectly serialized boolean; "
                         f"expected b'\\x00' or b'\\x01', got {repr(data)}")
    return b



from .__about__ import (
    __author__,  __license__, __summary__, __title__, __version__, __copyright__, __email__, __url__
)

from .capsule import Capsule
from .capsule_frag import CapsuleFrag, VerifiedCapsuleFrag
from .errors import VerificationError
from .key_frag import KeyFrag, VerifiedKeyFrag
from .keys import SecretKey, PublicKey, SecretKeyFactory
from .pre import encrypt, decrypt_original, decrypt_reencrypted, reencrypt, generate_kfrags
from .signing import Signature, Signer

__all__ = [
    "SecretKey",
    "PublicKey",
    "SecretKeyFactory",
    "Signature",
    "Signer",
    "Capsule",
    "KeyFrag",
    "VerifiedKeyFrag",
    "CapsuleFrag",
    "VerifiedCapsuleFrag",
    "VerificationError",
    "encrypt",
    "decrypt_original",
    "generate_kfrags",
    "reencrypt",
    "decrypt_reencrypted",
]



import os
from typing import Tuple

from .curve_scalar import CurveScalar
from .curve_point import CurvePoint
from .dem import kdf
from .hashing import Hash
from .serializable import Serializable, Deserializable


class SecretKey(Serializable, Deserializable):
    """
    Umbral secret (private) key.
    """

    def __init__(self, scalar_key: CurveScalar):
        self._scalar_key = scalar_key
        # Precached public key.
        # We are assuming here that there will be on average more
        # derivations of a public key from a secret key than secret key instantiations.
        self._public_key = PublicKey(CurvePoint.generator() * self._scalar_key)

    @classmethod
    def random(cls) -> 'SecretKey':
        """
        Generates a random secret key and returns it.
        """
        return cls(CurveScalar.random_nonzero())

    def public_key(self) -> 'PublicKey':
        """
        Returns the associated public key.
        """
        return self._public_key

    def __eq__(self, other):
        return self._scalar_key == other._scalar_key

    def __str__(self):
        return f"{self.__class__.__name__}:..."

    def __hash__(self):
        raise RuntimeError("Hashing secret objects is not secure")

    def secret_scalar(self) -> CurveScalar:
        return self._scalar_key

    @classmethod
    def serialized_size(cls):
        return CurveScalar.serialized_size()

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        return cls(CurveScalar._from_exact_bytes(data))

    def __bytes__(self) -> bytes:
        return bytes(self._scalar_key)


class PublicKey(Serializable, Deserializable):
    """
    Umbral public key.

    Created using :py:meth:`SecretKey.public_key`.
    """

    def __init__(self, point_key: CurvePoint):
        self._point_key = point_key

    def point(self) -> CurvePoint:
        return self._point_key

    @classmethod
    def serialized_size(cls):
        return CurvePoint.serialized_size()

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        return cls(CurvePoint._from_exact_bytes(data))

    def __bytes__(self) -> bytes:
        return bytes(self._point_key)

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"

    def __eq__(self, other):
        return self._point_key == other._point_key

    def __hash__(self) -> int:
        return hash((self.__class__, bytes(self)))


class SecretKeyFactory(Serializable, Deserializable):
    """
    This class handles keyring material for Umbral, by allowing deterministic
    derivation of :py:class:`SecretKey` objects based on labels.

    Don't use this key material directly as a key.
    """

    _KEY_SEED_SIZE = 64
    _DERIVED_KEY_SIZE = 64

    def __init__(self, key_seed: bytes):
        self.__key_seed = key_seed

    @classmethod
    def random(cls) -> 'SecretKeyFactory':
        """
        Creates a random factory.
        """
        return cls(os.urandom(cls._KEY_SEED_SIZE))

    def secret_key_by_label(self, label: bytes) -> SecretKey:
        """
        Creates a :py:class:`SecretKey` from the given label.
        """
        tag = b"KEY_DERIVATION/" + label
        key = kdf(self.__key_seed, self._DERIVED_KEY_SIZE, info=tag)

        digest = Hash(tag)
        digest.update(key)
        scalar_key = CurveScalar.from_digest(digest)

        return SecretKey(scalar_key)

    @classmethod
    def serialized_size(cls):
        return cls._KEY_SEED_SIZE

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        return cls(data)

    def __bytes__(self) -> bytes:
        return bytes(self.__key_seed)

    def __str__(self):
        return f"{self.__class__.__name__}:..."

    def __hash__(self):
        raise RuntimeError("Hashing secret objects is not secure")



from typing import TYPE_CHECKING, Optional, Iterable, Union, List, cast

from cryptography.hazmat.primitives import hashes

from .openssl import backend, ErrorInvalidCompressedPoint
from .curve import CURVE
from .curve_scalar import CurveScalar
from .curve_point import CurvePoint
from .serializable import Serializable, bool_bytes

if TYPE_CHECKING: # pragma: no cover
    from .key_frag import KeyFragID
    from .keys import PublicKey


class Hash:

    OUTPUT_SIZE = 32

    def __init__(self, dst: Optional[bytes] = None):
        self._backend_hash_algorithm = hashes.SHA256()
        self._hash = hashes.Hash(self._backend_hash_algorithm, backend=backend)

        if dst is not None:
            len_dst = len(dst).to_bytes(4, byteorder='big')
            self.update(len_dst + dst)

    def update(self, data: Union[bytes, Serializable]) -> None:
        self._hash.update(bytes(data))

    def finalize(self) -> bytes:
        return self._hash.finalize()


def hash_to_polynomial_arg(precursor: CurvePoint,
                           pubkey: CurvePoint,
                           dh_point: CurvePoint,
                           kfrag_id: 'KeyFragID',
                           ) -> CurveScalar:
    digest = Hash(b"POLYNOMIAL_ARG")
    digest.update(precursor)
    digest.update(pubkey)
    digest.update(dh_point)
    digest.update(kfrag_id)
    return CurveScalar.from_digest(digest)


def hash_capsule_points(e: CurvePoint, v: CurvePoint) -> CurveScalar:
    digest = Hash(b"CAPSULE_POINTS")
    digest.update(e)
    digest.update(v)
    return CurveScalar.from_digest(digest)


def hash_to_shared_secret(precursor: CurvePoint,
                          pubkey: CurvePoint,
                          dh_point: CurvePoint
                          ) -> CurveScalar:
    digest = Hash(b"SHARED_SECRET")
    digest.update(precursor)
    digest.update(pubkey)
    digest.update(dh_point)
    return CurveScalar.from_digest(digest)


def hash_to_cfrag_verification(points: Iterable[CurvePoint]) -> CurveScalar:
    digest = Hash(b"CFRAG_VERIFICATION")
    for point in points:
        digest.update(point)
    return CurveScalar.from_digest(digest)


def kfrag_signature_message(kfrag_id: 'KeyFragID',
                            commitment: CurvePoint,
                            precursor: CurvePoint,
                            maybe_delegating_pk: Optional['PublicKey'],
                            maybe_receiving_pk: Optional['PublicKey'],
                            ) -> bytes:

    # Have to convert to bytes manually because `mypy` is not smart enough to resolve types.

    delegating_part = ([bool_bytes(True), bytes(maybe_delegating_pk)]
                        if maybe_delegating_pk
                        else [bool_bytes(False)])
    cast(List[Serializable], delegating_part)

    receiving_part = ([bool_bytes(True), bytes(maybe_receiving_pk)]
                       if maybe_receiving_pk
                       else [bool_bytes(False)])

    components = ([bytes(kfrag_id), bytes(commitment), bytes(precursor)] +
                  delegating_part +
                  receiving_part)

    return b''.join(components)


def unsafe_hash_to_point(dst: bytes, data: bytes) -> CurvePoint:
    """
    Hashes arbitrary data into a valid EC point of the specified curve,
    using the try-and-increment method.

    WARNING: Do not use when the input data is secret, as this implementation is not
    in constant time, and hence, it is not safe with respect to timing attacks.
    """

    len_data = len(data).to_bytes(4, byteorder='big')
    data_with_len = len_data + data
    sign = b'\x02'

    # We use an internal 32-bit counter as additional input
    for i in range(2**32):
        ibytes = i.to_bytes(4, byteorder='big')
        digest = Hash(dst)
        digest.update(data_with_len + ibytes)
        point_data = digest.finalize()[:CURVE.field_element_size]

        compressed_point = sign + point_data

        try:
            return CurvePoint.from_bytes(compressed_point)
        except ErrorInvalidCompressedPoint:
            # If it is not a valid point, continue on
            pass

    # Only happens with probability 2^(-32)
    raise ValueError('Could not hash input into the curve') # pragma: no cover



from typing import Tuple, Optional, Sequence, List

from .capsule import Capsule
from .capsule_frag import VerifiedCapsuleFrag, CapsuleFrag
from .dem import DEM
from .keys import PublicKey, SecretKey
from .key_frag import VerifiedKeyFrag, KeyFrag, KeyFragBase
from .signing import Signer


def encrypt(delegating_pk: PublicKey, plaintext: bytes) -> Tuple[Capsule, bytes]:
    """
    Generates and encapsulates a symmetric key and uses it to encrypt the given plaintext.

    Returns the KEM Capsule and the ciphertext.
    """
    capsule, key_seed = Capsule.from_public_key(delegating_pk)
    dem = DEM(bytes(key_seed))
    ciphertext = dem.encrypt(plaintext, authenticated_data=bytes(capsule))
    return capsule, ciphertext


def decrypt_original(delegating_sk: SecretKey, capsule: Capsule, ciphertext: bytes) -> bytes:
    """
    Opens the capsule using the delegator's key used for encryption and gets what's inside.
    We hope that's a symmetric key, which we use to decrypt the ciphertext
    and return the resulting cleartext.
    """
    key_seed = capsule.open_original(delegating_sk)
    dem = DEM(bytes(key_seed))
    return dem.decrypt(ciphertext, authenticated_data=bytes(capsule))


def generate_kfrags(delegating_sk: SecretKey,
                    receiving_pk: PublicKey,
                    signer: Signer,
                    threshold: int,
                    num_kfrags: int,
                    sign_delegating_key: bool = True,
                    sign_receiving_key: bool = True,
                    ) -> List[VerifiedKeyFrag]:
    """
    Generates ``num_kfrags`` key fragments to pass to proxies for re-encryption.
    At least ``threshold`` of them will be needed for decryption.
    If ``sign_delegating_key`` or ``sign_receiving_key`` are ``True``,
    the corresponding keys will have to be provided to :py:meth:`KeyFrag.verify`.
    """

    base = KeyFragBase(delegating_sk, receiving_pk, signer, threshold)

    # Technically we could allow it, but what would be the use of these kfrags?
    if num_kfrags < threshold:
        raise ValueError(f"Creating less kfrags ({num_kfrags}) "
                         f"than threshold ({threshold}) makes them useless")

    kfrags = [KeyFrag.from_base(base, sign_delegating_key, sign_receiving_key)
              for _ in range(num_kfrags)]

    # Make them verified - we know they're good.
    return [VerifiedKeyFrag(kfrag) for kfrag in kfrags]


def reencrypt(capsule: Capsule, kfrag: VerifiedKeyFrag) -> VerifiedCapsuleFrag:
    """
    Creates a capsule fragment using the given key fragment.
    Capsule fragments can later be used to decrypt the ciphertext.
    """
    # We could let duck typing do its work,
    # but it's better to make a common error more understandable.
    if isinstance(kfrag, KeyFrag) and not isinstance(kfrag, VerifiedKeyFrag):
        raise TypeError("KeyFrag must be verified before reencryption")

    return VerifiedCapsuleFrag(CapsuleFrag.reencrypted(capsule, kfrag.kfrag))


def decrypt_reencrypted(receiving_sk: SecretKey,
                        delegating_pk: PublicKey,
                        capsule: Capsule,
                        verified_cfrags: Sequence[VerifiedCapsuleFrag],
                        ciphertext: bytes,
                        ) -> bytes:
    """
    Decrypts the ciphertext using the original capsule and the reencrypted capsule fragments.
    """
    # We could let duck typing do its work,
    # but it's better to make a common error more understandable.
    for cfrag in verified_cfrags:
        if isinstance(cfrag, CapsuleFrag) and not isinstance(cfrag, VerifiedCapsuleFrag):
            raise TypeError("All CapsuleFrags must be verified before decryption")

    cfrags = [vcfrag.cfrag for vcfrag in verified_cfrags]
    key_seed = capsule.open_reencrypted(receiving_sk, delegating_pk, cfrags)
    dem = DEM(bytes(key_seed))
    return dem.decrypt(ciphertext, authenticated_data=bytes(capsule))



from typing import Tuple

from cryptography.exceptions import InternalError, InvalidSignature
from cryptography.hazmat.backends.openssl import backend
from cryptography.hazmat.backends.openssl.ec import (_EllipticCurvePrivateKey,
                                                     _EllipticCurvePublicKey)
from cryptography.hazmat.primitives.asymmetric import utils
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA


BACKEND_LIB = backend._lib
BACKEND_FFI = backend._ffi


def tmp_bn_ctx():
    return backend._tmp_bn_ctx()


class Curve:
    """
    Acts as a container to store constant variables such as the OpenSSL
    curve_nid, the EC_GROUP struct, and the order of the curve.

    Contains a whitelist of supported elliptic curves used in pyUmbral.
    """

    _supported_curves = {
        714: 'secp256k1',
    }

    @staticmethod
    def _get_ec_group_by_curve_nid(nid: int):
        """
        Returns the group of a given curve via its OpenSSL nid. This must be freed
        after each use otherwise it leaks memory.
        """
        group = BACKEND_LIB.EC_GROUP_new_by_curve_name(nid)
        backend.openssl_assert(group != BACKEND_FFI.NULL)
        return group

    @staticmethod
    def _get_ec_order_by_group(ec_group):
        """
        Returns the order of a given curve via its OpenSSL EC_GROUP.
        """
        ec_order = _bn_new()
        with tmp_bn_ctx() as bn_ctx:
            res = BACKEND_LIB.EC_GROUP_get_order(ec_group, ec_order, bn_ctx)
            backend.openssl_assert(res == 1)
        return ec_order

    @staticmethod
    def _get_ec_generator_by_group(ec_group):
        """
        Returns the generator point of a given curve via its OpenSSL EC_GROUP.
        """
        generator = BACKEND_LIB.EC_GROUP_get0_generator(ec_group)
        backend.openssl_assert(generator != BACKEND_FFI.NULL)
        generator = BACKEND_FFI.gc(generator, BACKEND_LIB.EC_POINT_clear_free)

        return generator

    @staticmethod
    def _get_ec_group_degree(ec_group):
        """
        Returns the number of bits needed to represent the order of the finite
        field upon the curve is based.
        """
        return BACKEND_LIB.EC_GROUP_get_degree(ec_group)

    def __init__(self, nid: int):
        """
        Instantiates an OpenSSL curve with the provided curve_nid and derives
        the proper EC_GROUP struct and order. You can _only_ instantiate curves
        with supported nids (see `Curve.supported_curves`).
        """

        try:
            self.name = self._supported_curves[nid]
        except KeyError as e:
            raise NotImplementedError("Curve NID {} is not supported.".format(nid)) from e

        self.nid = nid

        self.ec_group = self._get_ec_group_by_curve_nid(self.nid)
        self.bn_order = self._get_ec_order_by_group(self.ec_group)
        self.point_generator = self._get_ec_generator_by_group(self.ec_group)

        size_in_bits = self._get_ec_group_degree(self.ec_group)
        self.field_element_size = (size_in_bits + 7) // 8

        self.scalar_size = _bn_size(self.bn_order)
        self.order = bn_to_int(self.bn_order)

    @classmethod
    def from_name(cls, name: str) -> 'Curve':
        """
        Alternate constructor to generate a curve instance by its name.

        Raises NotImplementedError if the name cannot be mapped to a known
        supported curve NID.
        """

        name = name.casefold()  # normalize

        for supported_nid, supported_name in cls._supported_curves.items():
            if name == supported_name:
                instance = cls(nid=supported_nid)
                break
        else:
            raise NotImplementedError(f"{name} is not supported curve name.")

        return instance

    def __eq__(self, other):
        return self.nid == other.nid

    def __str__(self):
        return "<OpenSSL Curve(nid={}, name={})>".format(self.nid, self.name)


#
# OpenSSL bignums
#


def _bn_new():
    """
    Returns a new and initialized OpenSSL BIGNUM.
    """
    new_bn = BACKEND_LIB.BN_new()
    backend.openssl_assert(new_bn != BACKEND_FFI.NULL)
    new_bn = BACKEND_FFI.gc(new_bn, BACKEND_LIB.BN_clear_free)

    # Always use constant time operations.
    BACKEND_LIB.BN_set_flags(new_bn, BACKEND_LIB.BN_FLG_CONSTTIME)
    return new_bn


def bn_is_normalized(check_bn, modulus):
    """
    Returns ``True`` if ``check_bn`` is in ``[0, modulus)``, ``False`` otherwise.
    """
    zero = backend._int_to_bn(0)
    zero = BACKEND_FFI.gc(zero, BACKEND_LIB.BN_clear_free)

    check_sign = BACKEND_LIB.BN_cmp(check_bn, zero)
    range_check = BACKEND_LIB.BN_cmp(check_bn, modulus)
    return check_sign in (0, 1) and range_check == -1


def bn_from_int(py_int: int, check_modulus=None):
    """
    Converts the given Python int to an OpenSSL BIGNUM. If ``modulus`` is
    provided, it will check if the Python integer is within ``[0, modulus)``.
    """
    conv_bn = backend._int_to_bn(py_int)
    conv_bn = BACKEND_FFI.gc(conv_bn, BACKEND_LIB.BN_clear_free)

    if check_modulus and not bn_is_normalized(conv_bn, check_modulus):
        raise ValueError(f"The Python integer given ({py_int}) is not under the provided modulus.")

    BACKEND_LIB.BN_set_flags(conv_bn, BACKEND_LIB.BN_FLG_CONSTTIME)
    return conv_bn


def bn_from_bytes(bytes_seq: bytes, check_modulus=None, apply_modulus=None):
    """
    Converts the given byte sequence to an OpenSSL BIGNUM.
    """
    bn = _bn_new()
    BACKEND_LIB.BN_bin2bn(bytes_seq, len(bytes_seq), bn)
    backend.openssl_assert(bn != BACKEND_FFI.NULL)

    if check_modulus and not bn_is_normalized(bn, check_modulus):
        raise ValueError(f"The integer encoded with given bytes ({repr(bytes_seq)}) "
                          "is not under the provided modulus.")

    if apply_modulus:
        bignum =_bn_new()
        with tmp_bn_ctx() as bn_ctx:
            res = BACKEND_LIB.BN_mod(bignum, bn, apply_modulus, bn_ctx)
            backend.openssl_assert(res == 1)
        return bignum

    return bn


def bn_to_bytes(bn, length: int):
    """
    Converts the given OpenSSL BIGNUM into a Python bytes sequence.
    If length is given, the return bytes will have such length.
    If the BIGNUM doesn't fit, it raises a ValueError.
    """

    # Sanity check, CurveScalar ensures it won't happen.
    bn_num_bytes = BACKEND_LIB.BN_num_bytes(bn)
    assert bn_num_bytes <= length, f"Input BIGNUM doesn't fit in {length} B"

    bin_ptr = BACKEND_FFI.new("unsigned char []", length)
    bin_len = BACKEND_LIB.BN_bn2bin(bn, bin_ptr)
    return bytes.rjust(BACKEND_FFI.buffer(bin_ptr, bin_len)[:], length, b'\0')


def bn_random_nonzero(modulus):

    one = BACKEND_LIB.BN_value_one()

    # TODO: in most cases, we want this number to be secret.
    # OpenSSL 1.1.1 has `BN_priv_rand_range()`, but it is not
    # currently exported by `cryptography`.
    # Use when available.

    # Calculate `modulus - 1`
    modulus_minus_1 = _bn_new()
    res = BACKEND_LIB.BN_sub(modulus_minus_1, modulus, one)
    backend.openssl_assert(res == 1)

    # Get a random in range `[0, modulus - 1)`
    new_rand_bn = _bn_new()
    res = BACKEND_LIB.BN_rand_range(new_rand_bn, modulus_minus_1)
    backend.openssl_assert(res == 1)

    # Turn it into a random in range `[1, modulus)`
    op_sum = _bn_new()
    res = BACKEND_LIB.BN_add(op_sum, new_rand_bn, one)
    backend.openssl_assert(res == 1)

    return op_sum


def _bn_size(bn):
    return BACKEND_LIB.BN_num_bytes(bn)


def bn_to_int(bn) -> int:
    return backend._bn_to_int(bn)


def bn_cmp(bn1, bn2):
    # -1 less than, 0 is equal to, 1 is greater than
    return BACKEND_LIB.BN_cmp(bn1, bn2)


def bn_one():
    return BACKEND_LIB.BN_value_one()


def bn_is_zero(bn):
    # No special function exported in the current backend, so this will have to do
    return bn_cmp(bn, bn_from_int(0)) == 0


def bn_invert(bn, modulus):
    with tmp_bn_ctx() as bn_ctx:
        inv = BACKEND_LIB.BN_mod_inverse(BACKEND_FFI.NULL, bn, modulus, bn_ctx)
        backend.openssl_assert(inv != BACKEND_FFI.NULL)
        inv = BACKEND_FFI.gc(inv, BACKEND_LIB.BN_clear_free)
    return inv


def bn_sub(bn1, bn2, modulus):
    diff = _bn_new()
    with tmp_bn_ctx() as bn_ctx:
        res = BACKEND_LIB.BN_mod_sub(diff, bn1, bn2, modulus, bn_ctx)
        backend.openssl_assert(res == 1)
    return diff


def bn_add(bn1, bn2, modulus):
    op_sum = _bn_new()
    with tmp_bn_ctx() as bn_ctx:
        res = BACKEND_LIB.BN_mod_add(op_sum, bn1, bn2, modulus, bn_ctx)
        backend.openssl_assert(res == 1)
    return op_sum


def bn_mul(bn1, bn2, modulus):
    product = _bn_new()
    with tmp_bn_ctx() as bn_ctx:
        res = BACKEND_LIB.BN_mod_mul(product, bn1, bn2, modulus, bn_ctx)
        backend.openssl_assert(res == 1)
    return product


def bn_to_privkey(curve: Curve, bn):

    ec_key = BACKEND_LIB.EC_KEY_new()
    backend.openssl_assert(ec_key != BACKEND_FFI.NULL)
    ec_key = BACKEND_FFI.gc(ec_key, BACKEND_LIB.EC_KEY_free)

    set_group_result = BACKEND_LIB.EC_KEY_set_group(ec_key, curve.ec_group)
    backend.openssl_assert(set_group_result == 1)

    set_privkey_result = BACKEND_LIB.EC_KEY_set_private_key(ec_key, bn)
    backend.openssl_assert(set_privkey_result == 1)

    evp_pkey = backend._ec_cdata_to_evp_pkey(ec_key)
    return _EllipticCurvePrivateKey(backend, ec_key, evp_pkey)


#
# OpenSSL EC points
#


def _point_new(ec_group):
    """
    Returns a new and initialized OpenSSL EC_POINT given the group of a curve.
    If __curve_nid is provided, it retrieves the group from the curve provided.
    """
    new_point = BACKEND_LIB.EC_POINT_new(ec_group)
    backend.openssl_assert(new_point != BACKEND_FFI.NULL)
    new_point = BACKEND_FFI.gc(new_point, BACKEND_LIB.EC_POINT_clear_free)

    return new_point


def point_to_affine_coords(curve: Curve, point) -> Tuple[int, int]:
    """
    Returns the affine coordinates of a given point on the provided ec_group.
    """
    affine_x = _bn_new()
    affine_y = _bn_new()

    try:
        with tmp_bn_ctx() as bn_ctx:
            res = BACKEND_LIB.EC_POINT_get_affine_coordinates_GFp(
                curve.ec_group, point, affine_x, affine_y, bn_ctx
            )
            backend.openssl_assert(res == 1)
    except InternalError as e:
        raise ValueError("Cannot get affine coordinates of an identity point")

    return bn_to_int(affine_x), bn_to_int(affine_y)


class ErrorInvalidCompressedPoint(Exception):
    pass


class ErrorInvalidPointEncoding(Exception):
    pass


def point_from_bytes(curve: Curve, data):
    point = _point_new(curve.ec_group)
    try:
        with tmp_bn_ctx() as bn_ctx:
            res = BACKEND_LIB.EC_POINT_oct2point(curve.ec_group, point, data, len(data), bn_ctx)
            backend.openssl_assert(res == 1)
    except InternalError as e:
        # We want to catch specific InternalExceptions.
        # https://github.com/openssl/openssl/blob/master/include/openssl/ecerr.h
        # There is also EC_R_POINT_IS_NOT_ON_CURVE (code 107),
        # but somehow it is never triggered during deserialization.
        if e.err_code[0].reason == 110: # EC_R_INVALID_COMPRESSED_POINT
            raise ErrorInvalidCompressedPoint from e
        if e.err_code[0].reason == 102: # EC_R_INVALID_ENCODING
            raise ErrorInvalidPointEncoding from e

        # Any other exception, we raise it.
        # (although at the moment I'm not sure what should one do to cause it)
        raise # pragma: no cover
    return point


def point_to_bytes_compressed(curve: Curve, point):
    point_conversion_form = BACKEND_LIB.POINT_CONVERSION_COMPRESSED

    size = curve.field_element_size + 1 # compressed point size

    bin_ptr = BACKEND_FFI.new("unsigned char[]", size)
    with tmp_bn_ctx() as bn_ctx:
        bin_len = BACKEND_LIB.EC_POINT_point2oct(
            curve.ec_group, point, point_conversion_form,
            bin_ptr, size, bn_ctx
        )
        backend.openssl_assert(bin_len != 0)

    return bytes(BACKEND_FFI.buffer(bin_ptr, bin_len)[:])


def point_eq(curve: Curve, point1, point2):
    with tmp_bn_ctx() as bn_ctx:
        is_equal = BACKEND_LIB.EC_POINT_cmp(curve.ec_group, point1, point2, bn_ctx)
        backend.openssl_assert(is_equal != -1)

    # 1 is not-equal, 0 is equal, -1 is error
    return is_equal == 0


def point_mul_bn(curve: Curve, point, bn):
    prod = _point_new(curve.ec_group)
    with tmp_bn_ctx() as bn_ctx:
        res = BACKEND_LIB.EC_POINT_mul(curve.ec_group, prod, BACKEND_FFI.NULL, point, bn, bn_ctx)
        backend.openssl_assert(res == 1)
    return prod


def point_add(curve: Curve, point1, point2):
    op_sum = _point_new(curve.ec_group)
    with tmp_bn_ctx() as bn_ctx:
        res = BACKEND_LIB.EC_POINT_add(curve.ec_group, op_sum, point1, point2, bn_ctx)
        backend.openssl_assert(res == 1)
    return op_sum


def point_neg(curve: Curve, point):
    inv = BACKEND_LIB.EC_POINT_dup(point, curve.ec_group)
    backend.openssl_assert(inv != BACKEND_FFI.NULL)
    inv = BACKEND_FFI.gc(inv, BACKEND_LIB.EC_POINT_clear_free)

    with tmp_bn_ctx() as bn_ctx:
        res = BACKEND_LIB.EC_POINT_invert(curve.ec_group, inv, bn_ctx)
        backend.openssl_assert(res == 1)

    return inv


def point_to_pubkey(curve: Curve, point):

    ec_key = BACKEND_LIB.EC_KEY_new()
    backend.openssl_assert(ec_key != BACKEND_FFI.NULL)
    ec_key = BACKEND_FFI.gc(ec_key, BACKEND_LIB.EC_KEY_free)

    set_group_result = BACKEND_LIB.EC_KEY_set_group(ec_key, curve.ec_group)
    backend.openssl_assert(set_group_result == 1)

    set_pubkey_result = BACKEND_LIB.EC_KEY_set_public_key(ec_key, point)
    backend.openssl_assert(set_pubkey_result == 1)

    evp_pkey = backend._ec_cdata_to_evp_pkey(ec_key)
    return _EllipticCurvePublicKey(backend, ec_key, evp_pkey)


#
# Signing
#

def ecdsa_sign(curve: Curve,
               secret_bn,
               prehashed_message: bytes,
               hash_algorithm
               ) -> Tuple[int, int]:
    signature_algorithm = ECDSA(utils.Prehashed(hash_algorithm))
    private_key = bn_to_privkey(curve, secret_bn)
    signature_der_bytes = private_key.sign(prehashed_message, signature_algorithm)
    r_int, s_int = utils.decode_dss_signature(signature_der_bytes)
    return r_int, s_int

def ecdsa_verify(curve: Curve, sig_r: int, sig_s: int, public_point,
                 prehashed_message: bytes, hash_algorithm) -> bool:
    signature_algorithm = ECDSA(utils.Prehashed(hash_algorithm))
    public_key = point_to_pubkey(curve, public_point)
    signature_der_bytes = utils.encode_dss_signature(sig_r, sig_s)

    try:
        public_key.verify(signature=signature_der_bytes,
                          data=prehashed_message,
                          signature_algorithm=signature_algorithm)
    except InvalidSignature:
        return False
    return True



from typing import Tuple

from . import openssl
from .curve import CURVE
from .curve_scalar import CurveScalar
from .serializable import Serializable, Deserializable


class CurvePoint(Serializable, Deserializable):
    """
    Represents an OpenSSL EC_POINT except more Pythonic.
    """

    def __init__(self, backend_point) -> None:
        self._backend_point = backend_point

    @classmethod
    def generator(cls) -> 'CurvePoint':
        return cls(CURVE.point_generator)

    @classmethod
    def random(cls) -> 'CurvePoint':
        """
        Returns a CurvePoint object with a cryptographically secure EC_POINT based
        on the provided curve.
        """
        return cls.generator() * CurveScalar.random_nonzero()

    def to_affine(self) -> Tuple[int, int]:
        """
        Returns a tuple of Python ints in the format of (x, y) that represents
        the point in the curve.
        """
        return openssl.point_to_affine_coords(CURVE, self._backend_point)

    @classmethod
    def serialized_size(cls):
        return CURVE.field_element_size + 1 # compressed point size

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        """
        Returns a CurvePoint object from the given byte data on the curve provided.
        """
        return cls(openssl.point_from_bytes(CURVE, data))

    def __bytes__(self) -> bytes:
        """
        Returns the CurvePoint serialized as bytes in the compressed form.
        """
        return openssl.point_to_bytes_compressed(CURVE, self._backend_point)

    def __eq__(self, other):
        """
        Compares two EC_POINTS for equality.
        """
        return openssl.point_eq(CURVE, self._backend_point, other._backend_point)

    def __mul__(self, other: CurveScalar) -> 'CurvePoint':
        """
        Performs an EC_POINT_mul on an EC_POINT and a BIGNUM.
        """
        return CurvePoint(openssl.point_mul_bn(CURVE, self._backend_point, other._backend_bignum))

    def __add__(self, other: 'CurvePoint') -> 'CurvePoint':
        """
        Performs an EC_POINT_add on two EC_POINTS.
        """
        return CurvePoint(openssl.point_add(CURVE, self._backend_point, other._backend_point))

    def __sub__(self, other: 'CurvePoint') -> 'CurvePoint':
        """
        Performs subtraction by adding the inverse of the `other` to the point.
        """
        return self + (-other)

    def __neg__(self) -> 'CurvePoint':
        """
        Computes the additive inverse of a CurvePoint, by performing an
        EC_POINT_invert on itself.
        """
        return CurvePoint(openssl.point_neg(CURVE, self._backend_point))



import os
from typing import Optional

from cryptography.hazmat.primitives.kdf.hkdf import HKDF
from cryptography.hazmat.primitives import hashes

import nacl
from nacl.bindings.crypto_aead import (
    crypto_aead_xchacha20poly1305_ietf_encrypt as xchacha_encrypt,
    crypto_aead_xchacha20poly1305_ietf_decrypt as xchacha_decrypt,
    crypto_aead_xchacha20poly1305_ietf_KEYBYTES as XCHACHA_KEY_SIZE,
    crypto_aead_xchacha20poly1305_ietf_NPUBBYTES as XCHACHA_NONCE_SIZE,
    crypto_aead_xchacha20poly1305_ietf_ABYTES as XCHACHA_TAG_SIZE,
    )

from . import openssl


def kdf(data: bytes,
        key_length: int,
        salt: Optional[bytes] = None,
        info: Optional[bytes] = None,
        ) -> bytes:

    hkdf = HKDF(algorithm=hashes.SHA256(),
                length=key_length,
                salt=salt,
                info=info,
                backend=openssl.backend)
    return hkdf.derive(data)


class DEM:

    KEY_SIZE = XCHACHA_KEY_SIZE
    NONCE_SIZE = XCHACHA_NONCE_SIZE
    TAG_SIZE = XCHACHA_TAG_SIZE

    def __init__(self,
                 key_material: bytes,
                 salt: Optional[bytes] = None,
                 info: Optional[bytes] = None,
                 ):
        self._key = kdf(key_material, self.KEY_SIZE, salt, info)

    def encrypt(self, plaintext: bytes, authenticated_data: bytes = b"") -> bytes:
        nonce = os.urandom(self.NONCE_SIZE)
        ciphertext = xchacha_encrypt(plaintext, authenticated_data, nonce, self._key)
        return nonce + ciphertext

    def decrypt(self, nonce_and_ciphertext: bytes, authenticated_data: bytes = b"") -> bytes:

        if len(nonce_and_ciphertext) < self.NONCE_SIZE:
            raise ValueError("The ciphertext must include the nonce")

        nonce = nonce_and_ciphertext[:self.NONCE_SIZE]
        ciphertext = nonce_and_ciphertext[self.NONCE_SIZE:]

        # Prevent an out of bounds error deep in NaCl
        if len(ciphertext) < self.TAG_SIZE:
            raise ValueError("The authentication tag is missing or malformed")

        try:
            return xchacha_decrypt(ciphertext, authenticated_data, nonce, self._key)
        except nacl.exceptions.CryptoError as e:
            raise ValueError("Decryption of ciphertext failed: "
                             "either someone tampered with the ciphertext or "
                             "you are using an incorrect decryption key.") from e



from typing import TYPE_CHECKING, Tuple, Sequence

from .curve_point import CurvePoint
from .curve_scalar import CurveScalar
from .hashing import hash_capsule_points, hash_to_polynomial_arg, hash_to_shared_secret
from .keys import PublicKey, SecretKey
from .serializable import Serializable, Deserializable
if TYPE_CHECKING: # pragma: no cover
    from .capsule_frag import CapsuleFrag


def lambda_coeff(xs: Sequence[CurveScalar], i: int) -> CurveScalar:
    res = CurveScalar.one()
    for j, xs_j in enumerate(xs):
        if j != i:
            inv_diff = (xs_j - xs[i]).invert()
            res = (res * xs_j) * inv_diff
    return res


class Capsule(Serializable, Deserializable):
    """
    Encapsulated symmetric key.
    """

    def __init__(self, point_e: CurvePoint, point_v: CurvePoint, signature: CurveScalar):
        self.point_e = point_e
        self.point_v = point_v
        self.signature = signature

    _COMPONENT_TYPES = CurvePoint, CurvePoint, CurveScalar
    _SERIALIZED_SIZE = sum(tp.serialized_size() for tp in _COMPONENT_TYPES)

    @classmethod
    def serialized_size(cls):
        return cls._SERIALIZED_SIZE

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        capsule = cls(*cls._split(data, *cls._COMPONENT_TYPES))
        if not capsule._verify():
            raise ValueError("Capsule self-verification failed. Serialized data may be damaged.")
        return capsule

    def __bytes__(self):
        return bytes(self.point_e) + bytes(self.point_v) + bytes(self.signature)

    @classmethod
    def from_public_key(cls, delegating_pk: PublicKey) -> Tuple['Capsule', CurvePoint]:
        g = CurvePoint.generator()

        priv_r = CurveScalar.random_nonzero()
        pub_r = g * priv_r

        priv_u = CurveScalar.random_nonzero()
        pub_u = g * priv_u

        h = hash_capsule_points(pub_r, pub_u)
        s = priv_u + (priv_r * h)

        shared_key = delegating_pk._point_key * (priv_r + priv_u)

        return cls(point_e=pub_r, point_v=pub_u, signature=s), shared_key

    def open_original(self, delegating_sk: SecretKey) -> CurvePoint:
        return (self.point_e + self.point_v) * delegating_sk.secret_scalar()

    def open_reencrypted(self,
                         receiving_sk: SecretKey,
                         delegating_pk: PublicKey,
                         cfrags: Sequence['CapsuleFrag'],
                         ) -> CurvePoint:

        if len(cfrags) == 0:
            raise ValueError("Empty CapsuleFrag sequence")

        precursor = cfrags[0].precursor

        if len(set(cfrags)) != len(cfrags):
            raise ValueError("Some of the CapsuleFrags are repeated")

        if not all(cfrag.precursor == precursor for cfrag in cfrags[1:]):
            raise ValueError("CapsuleFrags are not pairwise consistent")

        pub_key = receiving_sk.public_key().point()
        dh_point = precursor * receiving_sk.secret_scalar()

        # Combination of CFrags via Shamir's Secret Sharing reconstruction
        lc = [hash_to_polynomial_arg(precursor, pub_key, dh_point, cfrag.kfrag_id)
              for cfrag in cfrags]

        e_primes = []
        v_primes = []
        for i, cfrag in enumerate(cfrags):
            lambda_i = lambda_coeff(lc, i)
            e_primes.append(cfrag.point_e1 * lambda_i)
            v_primes.append(cfrag.point_v1 * lambda_i)
        e_prime = sum(e_primes[1:], e_primes[0])
        v_prime = sum(v_primes[1:], v_primes[0])

        # Secret value 'd' allows to make Umbral non-interactive
        d = hash_to_shared_secret(precursor, pub_key, dh_point)

        s = self.signature
        h = hash_capsule_points(self.point_e, self.point_v)

        orig_pub_key = delegating_pk.point()

        # TODO: check for d == 0? Or just let if fail?
        inv_d = d.invert()
        if orig_pub_key * (s * inv_d) != (e_prime * h) + v_prime:
            raise ValueError("Internal validation failed")

        return (e_prime + v_prime) * d

    def _components(self):
        return (self.point_e, self.point_v, self.signature)

    def _verify(self) -> bool:
        g = CurvePoint.generator()
        e, v, s = self._components()
        h = hash_capsule_points(e, v)
        return g * s == v + (e * h)

    def __eq__(self, other):
        return self._components() == other._components()

    def __hash__(self):
        return hash((self.__class__, bytes(self)))

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"



from . import openssl
from .curve import CURVE
from .curve_scalar import CurveScalar
from .hashing import Hash
from .keys import SecretKey, PublicKey
from .serializable import Serializable, Deserializable


def digest_for_signing(message: bytes) -> Hash:
    # Not using a DST here to make life easier for third-party verifiers
    digest = Hash()
    digest.update(message)
    return digest


class Signer:
    """
    An object possessing the capability to create signatures.
    For safety reasons serialization is prohibited.
    """

    def __init__(self, secret_key: SecretKey):
        self.__secret_key = secret_key

    def sign_digest(self, digest: Hash) -> 'Signature':

        secret_bn = self.__secret_key.secret_scalar()._backend_bignum
        r_int, s_int = openssl.ecdsa_sign(curve=CURVE,
                                          secret_bn=secret_bn,
                                          prehashed_message=digest.finalize(),
                                          hash_algorithm=digest._backend_hash_algorithm)

        # Normalize s. This is a non-malleability measure, which OpenSSL doesn't do.
        # See Bitcoin's BIP-0062 for more details:
        # https://github.com/bitcoin/bips/blob/master/bip-0062.mediawiki#Low_S_values_in_signatures

        # s is public, so no constant-timeness required here
        if s_int > (CURVE.order >> 1):
            s_int = CURVE.order - s_int

        # Already normalized, don't waste time
        r = CurveScalar.from_int(r_int, check_normalization=False)
        s = CurveScalar.from_int(s_int, check_normalization=False)

        return Signature(r, s)

    def sign(self, message: bytes) -> 'Signature':
        """
        Hashes and signs the message.
        """
        return self.sign_digest(digest_for_signing(message))

    def verifying_key(self) -> PublicKey:
        """
        Returns the public verification key corresponding to the secret key used for signing.
        """
        return self.__secret_key.public_key()

    def __str__(self):
        return f"{self.__class__.__name__}:..."

    def __hash__(self):
        raise RuntimeError(f"{self.__class__.__name__} objects do not support hashing")

    def __bytes__(self):
        raise RuntimeError(f"{self.__class__.__name__} objects do not support serialization")


class Signature(Serializable, Deserializable):
    """
    Wrapper for ECDSA signatures.
    """

    def __init__(self, r: CurveScalar, s: CurveScalar):
        self.r = r
        self.s = s

    def verify_digest(self, verifying_key: PublicKey, digest: Hash) -> bool:
        return openssl.ecdsa_verify(curve=CURVE,
                                    sig_r=int(self.r),
                                    sig_s=int(self.s),
                                    public_point=verifying_key.point()._backend_point,
                                    prehashed_message=digest.finalize(),
                                    hash_algorithm=digest._backend_hash_algorithm)

    def verify(self, verifying_key: PublicKey, message: bytes) -> bool:
        """
        Returns ``True`` if the ``message`` was signed by someone possessing the secret counterpart
        to ``verifying_key``.
        """
        digest = digest_for_signing(message)
        return self.verify_digest(verifying_key, digest)

    @classmethod
    def serialized_size(cls):
        return CurveScalar.serialized_size() * 2

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        return cls(*cls._split(data, CurveScalar, CurveScalar))

    def __bytes__(self):
        return bytes(self.r) + bytes(self.s)

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"

    def __eq__(self, other):
        return self.r == other.r and self.s == other.s

    def __hash__(self) -> int:
        return hash((self.__class__, bytes(self)))



from typing import TYPE_CHECKING, Union, Tuple

from . import openssl
from .curve import CURVE
from .serializable import Serializable, Deserializable
if TYPE_CHECKING: # pragma: no cover
    from .hashing import Hash


class CurveScalar(Serializable, Deserializable):
    """
    Represents an OpenSSL Bignum modulo the order of a curve. Some of these
    operations will only work with prime numbers.

    By default, the underlying OpenSSL BIGNUM has BN_FLG_CONSTTIME set for
    constant time operations.
    """

    def __init__(self, backend_bignum):
        self._backend_bignum = backend_bignum

    @classmethod
    def random_nonzero(cls) -> 'CurveScalar':
        """
        Returns a CurveScalar object with a cryptographically secure OpenSSL BIGNUM.
        """
        return cls(openssl.bn_random_nonzero(CURVE.bn_order))

    @classmethod
    def from_int(cls, num: int, check_normalization: bool = True) -> 'CurveScalar':
        """
        Returns a CurveScalar object from a given integer on a curve.
        """
        modulus = CURVE.bn_order if check_normalization else None
        conv_bn = openssl.bn_from_int(num, check_modulus=modulus)
        return cls(conv_bn)

    @classmethod
    def from_digest(cls, digest: 'Hash') -> 'CurveScalar':
        # TODO (#39): this is used in Umbral scheme itself,
        # and needs to be able to return a guaranteed nonzero scalar.
        # Currently just matching what we have in rust-umbral
        # (taking bytes modulo curve order).
        # Can produce zeros!
        return cls(openssl.bn_from_bytes(digest.finalize(), apply_modulus=CURVE.bn_order))

    @classmethod
    def serialized_size(cls):
        return CURVE.scalar_size

    @classmethod
    def _from_exact_bytes(cls, data: bytes):
        return cls(openssl.bn_from_bytes(data, check_modulus=CURVE.bn_order))

    def __bytes__(self) -> bytes:
        """
        Returns the CurveScalar as bytes.
        """
        return openssl.bn_to_bytes(self._backend_bignum, CURVE.scalar_size)

    def __int__(self) -> int:
        """
        Converts the CurveScalar to a Python int.
        """
        return openssl.bn_to_int(self._backend_bignum)

    def __eq__(self, other) -> bool:
        """
        Compares the two BIGNUMS or int.
        """
        if isinstance(other, int):
            other = CurveScalar.from_int(other)
        return openssl.bn_cmp(self._backend_bignum, other._backend_bignum) == 0

    @classmethod
    def one(cls):
        return cls(openssl.bn_one())

    def is_zero(self):
        return openssl.bn_is_zero(self._backend_bignum)

    def __mul__(self, other: Union[int, 'CurveScalar']) -> 'CurveScalar':
        """
        Performs a BN_mod_mul between two BIGNUMS.
        """
        if isinstance(other, int):
            other = CurveScalar.from_int(other)
        return CurveScalar(openssl.bn_mul(self._backend_bignum,
                                          other._backend_bignum,
                                          CURVE.bn_order))

    def __add__(self, other : Union[int, 'CurveScalar']) -> 'CurveScalar':
        """
        Performs a BN_mod_add on two BIGNUMs.
        """
        if isinstance(other, int):
            other = CurveScalar.from_int(other)
        return CurveScalar(openssl.bn_add(self._backend_bignum,
                                          other._backend_bignum,
                                          CURVE.bn_order))

    def __sub__(self, other : Union[int, 'CurveScalar']) -> 'CurveScalar':
        """
        Performs a BN_mod_sub on two BIGNUMS.
        """
        if isinstance(other, int):
            other = CurveScalar.from_int(other)
        return CurveScalar(openssl.bn_sub(self._backend_bignum,
                                          other._backend_bignum,
                                          CURVE.bn_order))

    def invert(self) -> 'CurveScalar':
        """
        Performs a BN_mod_inverse.
        WARNING: Only in constant time if BN_FLG_CONSTTIME is set on the BN.
        """
        return CurveScalar(openssl.bn_invert(self._backend_bignum, CURVE.bn_order))



import os
from typing import List, Optional, Tuple, Type

from .curve_point import CurvePoint
from .curve_scalar import CurveScalar
from .errors import VerificationError
from .hashing import hash_to_shared_secret, kfrag_signature_message, hash_to_polynomial_arg
from .keys import PublicKey, SecretKey
from .params import PARAMETERS
from .serializable import (
    Serializable, Deserializable, HasSerializedSize,
    bool_bytes, bool_serialized_size)
from .signing import Signature, Signer


class KeyFragID(Serializable, Deserializable):

    __SIZE = 32

    def __init__(self, id_: bytes):
        self._id = id_

    def __eq__(self, other):
        return self._id == other._id

    @classmethod
    def random(cls) -> 'KeyFragID':
        return cls(os.urandom(cls.__SIZE))

    @classmethod
    def serialized_size(cls):
        return cls.__SIZE

    @classmethod
    def _from_exact_bytes(cls, data):
        return cls(data)

    def __bytes__(self):
        return self._id


class KeyFragProof(Serializable, Deserializable):

    @classmethod
    def from_base(cls,
                  base: 'KeyFragBase',
                  kfrag_id: KeyFragID,
                  kfrag_key: CurveScalar,
                  sign_delegating_key: bool,
                  sign_receiving_key: bool,
                  ) -> 'KeyFragProof':

        params = PARAMETERS

        kfrag_precursor = base.precursor
        signer = base.signer
        delegating_pk = base.delegating_pk
        receiving_pk = base.receiving_pk

        commitment = params.u * kfrag_key

        message_for_receiver = kfrag_signature_message(kfrag_id=kfrag_id,
                                                       commitment=commitment,
                                                       precursor=kfrag_precursor,
                                                       maybe_delegating_pk=delegating_pk,
                                                       maybe_receiving_pk=receiving_pk,
                                                       )
        signature_for_receiver = signer.sign(message_for_receiver)

        maybe_delegating_pk = delegating_pk if sign_delegating_key else None
        maybe_receiving_pk = receiving_pk if sign_receiving_key else None
        message_for_proxy = kfrag_signature_message(kfrag_id=kfrag_id,
                                                    commitment=commitment,
                                                    precursor=kfrag_precursor,
                                                    maybe_delegating_pk=maybe_delegating_pk,
                                                    maybe_receiving_pk=maybe_receiving_pk,
                                                    )
        signature_for_proxy = signer.sign(message_for_proxy)

        return cls(commitment,
                   signature_for_proxy,
                   signature_for_receiver,
                   sign_delegating_key,
                   sign_receiving_key)

    def __init__(self,
                 commitment: CurvePoint,
                 signature_for_proxy: Signature,
                 signature_for_receiver: Signature,
                 delegating_key_signed: bool,
                 receiving_key_signed: bool
                 ):

        self.commitment = commitment
        self.signature_for_proxy = signature_for_proxy
        self.signature_for_receiver = signature_for_receiver
        self.delegating_key_signed = delegating_key_signed
        self.receiving_key_signed = receiving_key_signed

    def _components(self):
        return (self.commitment,
                self.signature_for_proxy,
                self.signature_for_receiver,
                self.delegating_key_signed,
                self.receiving_key_signed)

    def __eq__(self, other):
        return self._components() == other._components()

    _SERIALIZED_SIZE = (CurvePoint.serialized_size() +
                        Signature.serialized_size() * 2 +
                        bool_serialized_size() * 2)

    @classmethod
    def serialized_size(cls):
        return cls._SERIALIZED_SIZE

    @classmethod
    def _from_exact_bytes(cls, data):
        types = [CurvePoint, Signature, Signature, bool, bool]
        return cls(*cls._split(data, *types))

    def __bytes__(self):
        return (bytes(self.commitment) +
                bytes(self.signature_for_proxy) +
                bytes(self.signature_for_receiver) +
                bool_bytes(self.delegating_key_signed) +
                bool_bytes(self.receiving_key_signed)
                )


# Coefficients of the generating polynomial
def poly_eval(coeffs: List[CurveScalar], x: CurveScalar) -> CurveScalar:
    result = coeffs[-1]
    for coeff in reversed(coeffs[:-1]):
        result = (result * x) + coeff
    return result


class KeyFrag(Serializable, Deserializable):
    """
    A signed fragment of the delegating key.
    """

    def __init__(self,
                 id_: KeyFragID,
                 key: CurveScalar,
                 precursor: CurvePoint,
                 proof: KeyFragProof):
        self.id = id_
        self.key = key
        self.precursor = precursor
        self.proof = proof

    def _components(self):
        return self.id, self.key, self.precursor, self.proof

    _COMPONENT_TYPES: Tuple[Type[HasSerializedSize], ...] = (
        KeyFragID, CurveScalar, CurvePoint, KeyFragProof)
    _SERIALIZED_SIZE = sum(tp.serialized_size() for tp in _COMPONENT_TYPES)

    @classmethod
    def serialized_size(cls):
        return cls._SERIALIZED_SIZE

    @classmethod
    def _from_exact_bytes(cls, data):
        return cls(*cls._split(data, *cls._COMPONENT_TYPES))

    def __bytes__(self):
        return b''.join(bytes(comp) for comp in self._components())

    def __eq__(self, other):
        return self._components() == other._components()

    def __hash__(self):
        return hash((self.__class__, bytes(self)))

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"

    @classmethod
    def from_base(cls,
                  base: 'KeyFragBase',
                  sign_delegating_key: bool,
                  sign_receiving_key: bool,
                  ) -> 'KeyFrag':

        kfrag_id = KeyFragID.random()

        # The index of the re-encryption key share (which in Shamir's Secret
        # Sharing corresponds to x in the tuple (x, f(x)), with f being the
        # generating polynomial), is used to prevent reconstruction of the
        # re-encryption key without Bob's intervention
        share_index = hash_to_polynomial_arg(base.precursor,
                                             base.receiving_pk.point(),
                                             base.dh_point,
                                             kfrag_id,
                                             )

        # The re-encryption key share is the result of evaluating the generating
        # polynomial for the index value
        rk = poly_eval(base.coefficients, share_index)

        proof = KeyFragProof.from_base(base,
                                       kfrag_id,
                                       rk,
                                       sign_delegating_key,
                                       sign_receiving_key,
                                       )

        return cls(kfrag_id, rk, base.precursor, proof)

    def verify(self,
               verifying_pk: PublicKey,
               delegating_pk: Optional[PublicKey] = None,
               receiving_pk: Optional[PublicKey] = None,
               ) -> 'VerifiedKeyFrag':
        """
        Verifies the validity of this fragment.

        If the delegating and/or receiving key were not signed in :py:func:`generate_kfrags`,
        but are given to this function, they are ignored.
        """

        u = PARAMETERS.u

        kfrag_id = self.id
        key = self.key
        commitment = self.proof.commitment
        precursor = self.precursor

        if commitment != u * key:
            raise VerificationError("Invalid kfrag commitment")

        if self.proof.delegating_key_signed and not bool(delegating_pk):
            raise VerificationError("A signature of a delegating key was included in this kfrag, "
                                    "but the key is not provided")

        if self.proof.receiving_key_signed and not bool(receiving_pk):
            raise VerificationError("A signature of a receiving key was included in this kfrag, "
                                    "but the key is not provided")

        delegating_pk = delegating_pk if self.proof.delegating_key_signed else None
        receiving_pk = receiving_pk if self.proof.receiving_key_signed else None
        kfrag_message = kfrag_signature_message(kfrag_id=kfrag_id,
                                                commitment=commitment,
                                                precursor=precursor,
                                                maybe_delegating_pk=delegating_pk,
                                                maybe_receiving_pk=receiving_pk)
        if not self.proof.signature_for_proxy.verify(verifying_pk, kfrag_message):
            raise VerificationError("Failed to verify the kfrag signature")

        return VerifiedKeyFrag(self)


class VerifiedKeyFrag(Serializable):
    """
    Verified kfrag, good for reencryption.
    Can be cast to ``bytes``, but cannot be deserialized from bytes directly.
    It can only be obtained from :py:meth:`KeyFrag.verify`.
    """

    def __init__(self, kfrag: KeyFrag):
        self.kfrag = kfrag

    def __bytes__(self):
        return bytes(self.kfrag)

    @classmethod
    def serialized_size(cls):
        return KeyFrag.serialized_size()

    @classmethod
    def from_verified_bytes(cls, data) -> 'VerifiedKeyFrag':
        """
        Restores a verified keyfrag directly from serialized bytes,
        skipping :py:meth:`KeyFrag.verify` call.

        Intended for internal storage;
        make sure that the bytes come from a trusted source.
        """
        kfrag = KeyFrag.from_bytes(data)
        return cls(kfrag)

    def __eq__(self, other):
        return self.kfrag == other.kfrag

    def __hash__(self):
        return hash((self.__class__, bytes(self)))

    def __str__(self):
        return f"{self.__class__.__name__}:{bytes(self).hex()[:16]}"


class KeyFragBase:

    def __init__(self,
                 delegating_sk: SecretKey,
                 receiving_pk: PublicKey,
                 signer: Signer,
                 threshold: int,
                 ):

        if threshold <= 0:
            raise ValueError(f"`threshold` must be larger than 0 (given: {threshold})")

        g = CurvePoint.generator()

        delegating_pk = delegating_sk.public_key()

        receiving_pk_point = receiving_pk.point()

        while True:
            # The precursor point is used as an ephemeral public key in a DH key exchange,
            # and the resulting shared secret 'dh_point' is used to derive other secret values
            private_precursor = CurveScalar.random_nonzero()
            precursor = g * private_precursor

            dh_point = receiving_pk_point * private_precursor

            # Secret value 'd' allows to make Umbral non-interactive
            d = hash_to_shared_secret(precursor, receiving_pk_point, dh_point)

            # At the moment we cannot statically ensure `d` is not zero,
            # but we need it to be non-zero for the algorithm to work.
            if not d.is_zero():
                break

        # Coefficients of the generating polynomial
        # `invert()` is guaranteed to work because `d` is nonzero.
        coefficients = [
            delegating_sk.secret_scalar() * d.invert(),
            *[CurveScalar.random_nonzero() for _ in range(threshold-1)]]

        self.signer = signer
        self.precursor = precursor
        self.dh_point = dh_point
        self.delegating_pk = delegating_pk
        self.receiving_pk = receiving_pk
        self.coefficients = coefficients



from __future__ import absolute_import, division, print_function

__all__ = [
    "__title__", "__summary__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__", "__url__"
]

__title__ = "umbral"

__url__ = "https://github.com/nucypher/pyUmbral"

__summary__ = 'NuCypher\'s Umbral Proxy Re-Encryption Implementation'

__version__ = "0.2.0"

__author__ = "NuCypher"

__email__ = "dev@nucypher.com"

__license__ = "GNU General Public License, Version 3"

__copyright__ = 'Copyright (C) 2019 NuCypher'
